import json

from django import forms, get_version
from django.conf import settings
from django.forms.renderers import get_default_renderer
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import get_language

if get_version() >= "4.0":
    from django.utils.translation import gettext_lazy as _
else:
    from django.utils.translation import ugettext_lazy as _
from django.forms.utils import ErrorList

DEFAULT_CONFIG = {
    "toolbar": ["heading", "|", "bold", "italic"],
}


class CKEditor5Widget(forms.Widget):
    template_name = "hs_django_ckeditor/widget.html"

    def __init__(self, config_name="default", attrs=None):
        self._config_errors = []
        self.config = DEFAULT_CONFIG.copy()
        try:
            configs = getattr(settings, "CKEDITOR_CONFIGS")
            try:
                self.config.update(configs[config_name])
            except (TypeError, KeyError, ValueError) as ex:
                self._config_errors.append(self.format_error(ex))
        except AttributeError as ex:
            self._config_errors.append(self.format_error(ex))

        default_attrs = {"class": "hs_django_ckeditor"}
        if attrs:
            default_attrs.update(attrs)
        super().__init__(default_attrs)

    def format_error(self, ex):
        return "{} {}".format(
            _("Check the correct settings.CKEDITOR_CONFIGS "),
            str(ex),
        )

    class Media:
        css = {
            "all": [
                "hs_django_ckeditor/dist/styles.css",
            ],
        }
        custom_css = getattr(settings, "CKEDITOR_CUSTOM_CSS", None)
        if custom_css:
            css["all"].append(custom_css)
        js = ["hs_django_ckeditor/dist/bundle.js"]
        configs = getattr(settings, "CKEDITOR_CONFIGS", None)
        if configs is not None:
            for config in configs:
                language = configs[config].get("language")
                if language:
                    languages = []
                    if isinstance(language, dict) and language.get("ui"):
                        language = language.get("ui")
                    elif isinstance(language, str):
                        languages.append(language)
                    elif isinstance(language, list):
                        languages = language
                    for lang in languages:
                        if lang != "en":
                            js += [f"hs_django_ckeditor/dist/translations/{lang}.js"]

    def render(self, name, value, attrs=None, renderer=None):
        context = super().get_context(name, value, attrs)
        use_language = getattr(settings, "CKEDITOR_USER_LANGUAGE", False)
        if use_language:
            language = get_language().lower()
            if language:
                self.config["language"] = language

        if renderer is None:
            renderer = get_default_renderer()

        context["config"] = self.config
        context["script_id"] = "{}{}".format(attrs["id"], "_script")
        context["upload_url"] = reverse(
            getattr(
                settings,
                "CK_EDITOR_5_UPLOAD_FILE_VIEW_NAME",
                "ck_editor_5_upload_file",
            ),
        )
        context["upload_file_types"] = json.dumps(
            getattr(
                settings,
                "CKEDITOR_UPLOAD_FILE_TYPES",
                ["jpeg", "png", "gif", "bmp", "webp", "tiff"],
            ),
        )
        context["csrf_cookie_name"] = settings.CSRF_COOKIE_NAME
        if self._config_errors:
            context["errors"] = ErrorList(self._config_errors)

        return mark_safe(renderer.render(self.template_name, context))



from django import get_version
from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible

if get_version() >= "4.0":
    from django.utils.translation import gettext_lazy as _
else:
    from django.utils.translation import ugettext_lazy as _


@deconstructible()
class FileMaxSizeValidator:
    """Validate that a file is not bigger than max_size mb, otherwise raise ValidationError.
    If zero is passed for max_size any file size is allowed.
    """

    message = _("File should be at most %(max_size)s MB.")
    code = "invalid_size"

    def __init__(self, max_size):
        self.max_size = max_size * 1024 * 1024
        self.orig_max_size = max_size

    def __call__(self, value):
        if value.size > self.max_size > 0:
            raise ValidationError(
                self.message,
                code=self.code,
                params={
                    "max_size": self.orig_max_size,
                },
            )

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.max_size == other.max_size
            and self.message == other.message
            and self.code == other.code
        )



from PIL import Image
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.utils.module_loading import import_string

from hs_django_ckeditor.exceptions import NoImageException


def get_storage_class():
    storage_setting = getattr(settings, "CKEDITOR_FILE_STORAGE", None)
    default_storage_setting = getattr(settings, "DEFAULT_FILE_STORAGE", None)
    storages_setting = getattr(settings, "STORAGES", {})
    default_storage_name = storages_setting.get("default", {}).get("BACKEND")

    if storage_setting:
        return import_string(storage_setting)
    elif default_storage_setting:
        try:
            return import_string(default_storage_setting)
        except ImportError:
            error_msg = f"Invalid default storage class: {default_storage_setting}"
            raise ImproperlyConfigured(error_msg)
    elif default_storage_name:
        try:
            return import_string(default_storage_name)
        except ImportError:
            error_msg = f"Invalid default storage class: {default_storage_name}"
            raise ImproperlyConfigured(error_msg)
    else:
        error_msg = (
            "Either CKEDITOR_FILE_STORAGE, DEFAULT_FILE_STORAGE, "
            "or STORAGES['default'] setting is required."
        )
        raise ImproperlyConfigured(error_msg)


def image_verify(f):
    try:
        Image.open(f).verify()
    except OSError:
        raise NoImageException


def handle_uploaded_file(f):
    storage = get_storage_class()
    fs = storage()
    filename = fs.save(f.name, f)
    return fs.url(filename)



from django.apps import AppConfig


class DjangoCkeditor5Config(AppConfig):
    name = "hs_django_ckeditor"
    verbose_name = "Django CkEditor"

    def ready(self):
        from . import signals  # noqa: F401



import os
import re

from django.db.models.signals import pre_delete, pre_save
from django.dispatch import receiver

from hs_django_ckeditor.fields import CKEditor5Field
from hs_django_ckeditor.storage_utils import get_storage_class

import logging

logger = logging.getLogger(__name__)

__all__ = [
    "cleanup_ckeditor_images_on_delete",
    "cleanup_unused_ckeditor_images_on_update",
]


def extract_image_paths(html_content):
    """
    Extracts image paths from the HTML content stored in CKEditor5Field.
    """
    if not html_content:
        return []
    return re.findall(r'<img[^>]+src=["\'](.*?)["\']', html_content)


def delete_images(storage, image_paths):
    """
    Deletes images from disk based on the provided paths.
    """
    for img_path in image_paths:
        file_name = os.path.basename(img_path)
        abs_path = os.path.join(storage.location, file_name)
        if os.path.exists(abs_path):
            os.remove(abs_path)


@receiver(pre_delete)
def cleanup_ckeditor_images_on_delete(sender, instance, **kwargs):
    """
    Removes images from disk when an object is deleted.
    If an error occurs, it is logged, but the deletion process continues.
    """
    try:
        storage = get_storage_class()
        if not storage:
            return

        # Find CKEditor5Field dynamically
        images_to_delete = []
        for field in instance._meta.fields:
            if isinstance(field, CKEditor5Field):
                images_to_delete.extend(
                    extract_image_paths(getattr(instance, field.name, ""))
                )

        try:
            delete_images(storage, images_to_delete)
        except Exception as e:
            logger.warning(f"Failed to delete images on object deletion: {e}")
    except Exception as e:
        logger.error(f"Error in cleanup_ckeditor_images_on_delete: {e}")


@receiver(pre_save)
def cleanup_unused_ckeditor_images_on_update(sender, instance, **kwargs):
    """
    Removes unused images when an object is updated.
    If any unexpected error occurs, it will be logged, but the deletion process won't break the update.
    """
    try:
        storage = get_storage_class()
        if not storage:
            return

        for field in instance._meta.fields:
            if isinstance(field, CKEditor5Field):
                try:
                    old_instance = sender.objects.get(pk=instance.pk)
                    old_images = set(
                        extract_image_paths(getattr(old_instance, field.name, ""))
                    )
                except sender.DoesNotExist:
                    old_images = set()

                new_images = set(extract_image_paths(getattr(instance, field.name, "")))
                unused_images = old_images - new_images

                try:
                    delete_images(storage, unused_images)
                except Exception as e:
                    logger.warning(f"Failed to delete unused images: {e}")
    except Exception as e:
        logger.error(f"Error in cleanup_unused_ckeditor_images_on_update: {e}")
